import numpy as np

def make_sim_data(n = 500, m = 10, alpha = 4.0, omega = 2, continuous_z = False, include_intercept = True, sample_xi = True, missing_data_prob = None):
    '''
    Generate simulated data from the Indian Buffet Process using the
    stick-breaking construction.
    
    Parameters
    ----------
    n: int, optional
        Number of observations to simulate.
    m: int, optional
        Number of observed variables.
    alpha: float, optional
        Parameter of the Indian buffet process.
    omega: float, optional
        Prior precision of A (and intercept if it is sampled).
    continuous_z: logical, optional
        If True then z is a product of normal and Bernoulli distributions.
        If False then only Bernoulli distributions are used to sample z.
    include_intercept: logical, optional
        If True then an intercept term is sampled from N(0, 1/sqrt(omega)).
        If False then no intercept term is included.
    sample_xi: logical, optional
        If True then xi (the observed variable precision) is sampled from uniform(0.8, 1.2).
        If False then xi is set to 1.
    missing_data_prob: float or None, optional
        If None (the default) then no data are missing.  If specified then this is the
        independent probability of each observed variable being missing in each observation
        (up to a maximum of 2 missing variables per observation). 
    '''
    # --sample latent features (IBP stick-breaking construction)--
    rng = np.random.default_rng()
    v = np.array([rng.beta(alpha, 1)])
    pi = np.array([v])
    # keep sampling feature probabilities (pi) until they become 
    # negligibly small and we have at least 8 of them
    while (pi[-1] > 0.01) or (len(pi) < 8):
        v = np.append(v, rng.beta(alpha, 1))
        pi = np.append(pi, v[-1] * pi[-1])
    p = len(pi) # true number of non-negligible latent features
    z = np.zeros([p, n])
    if continuous_z:
        for i in range(n):
            z[:, i] = rng.binomial(1, pi)
    else:
        for i in range(n):
            z[:, i] = rng.binomial(1, pi)*rng.normal(0, 1, size = p)

    # --sample loading matrix--
    A = rng.normal(0, 1/np.sqrt(omega), [m, p])

    # --sample intercept terms--
    if include_intercept:
        intercept = rng.normal(0, 1/np.sqrt(omega), m)
    else:
        intercept = np.zeros(m)
        
    # --sample observed variable precision--
    if sample_xi:
        xi = rng.uniform(0.8, 1.2, m) # noise standard deviation
    else:
        xi = np.ones(m)

    # --sample observations--
    x = np.zeros([m, n])
    for j in range(m):
        x[j, :] = rng.normal(A[j, :]@z + intercept[j], 1/np.sqrt(xi[j]))
        
    # --determine which data are missing, if any--
    if not missing_data_prob is None:
        for i in range(n):
            missing_count = 0
            for j in range(m):
                if rng.binomial(1, missing_data_prob) == 1:
                    x[j,i] = np.nan
                    missing_count += 1
                if missing_count >= 2:
                    break
    
    return {'x': x, 'A': A, 'intercept': intercept, 'xi': xi, 'z': z, 'v': v, 'pi': pi}


import numpy as np
import pandas as pd
from scipy.special import betaln, gammaln

# These classes are somewhat similar in spirit to this: https://pythonhosted.org/infpy/infpy.exp.html
# I used 1-indexing in the doc, so I'm using pseudo 1-indexing here to avoid mistakes in transcribing the math into Python.

class distribution:
        
    def tau_prime(self, y):
        T = self.T(y)
        n = y.shape[0]
        return [self.tau[j] + np.sum(T[j], axis = -1) for j in range(self.p)] + [self.tau[self.p] + n]
    
    def weighted_tau_prime(self, y, weights):
        '''
        Notes
        -----
        weights represents a weighting factor (should be from 0 to 1) for each 
        observation.
        '''
        T = self.T(y)
        n = np.sum(weights)
        return [self.tau[j] + np.sum(weights*T[j], axis = -1) for j in range(self.p)] + [self.tau[self.p] + n]    
    
    def update(self, y, weights = None):
        '''
        Perform conjugate prior updating.
        '''
        if weights is None:
            self.tau = self.tau_prime(y)
        else:
            self.tau = self.weighted_tau_prime(y, weights)
        self.hpar = self.tau_to_hpar(self.tau)
    
    def log_density(self, theta):
        '''
        Compute the log density of theta.
        '''
        return np.sum([np.sum(self.tau[j]*theta[j]) for j in range(self.p)]) - self.tau[self.p]*self.g(theta) - self.h(self.tau)
        
    def density(self, theta):
        return np.exp(self.log_density(theta))
        
    #def log_predictive(self, y):
    #    '''
    #    Compute the log predictive distribution of data.
    #    '''
        # ADD THIS
        
    #def predictive(self, y):
    #    return np.exp(self.log_predictive(y))
    
    def log_marginal(self, y):
        '''
        Compute the log of the marginal distribution of data.
        '''
        return -np.sum(self.f(y)) - self.h(self.tau) + self.h(self.tau_prime(y))            
        
    def weighted_log_marginal(self, y, weights):
        return -np.sum(weights*self.f(y)) - self.h(self.tau) + self.h(self.weighted_tau_prime(y, weights))
    
    def frac_log_marginal(self, y, frac):
        n = y.shape[0]
        return self.weighted_log_marginal(y = y, weights = frac*np.ones(n))
    
    def marginal(self, y):
        return np.exp(self.log_marginal(y))

class multi_group_distribution:
    '''
    This assumes no shared parameter (e.g. variance/precision) across groups.
    
    z gives 1-hot encoding of each observation's group (row = group, column = observation).
    
    n_t is the number of groups.
    
    I should revise __init__ in the future to allow the groups to have different starting hyperparameters.
    '''
    
    def __init__(self, base_dist, hpar, n_t):
        self.n_t = n_t
        self.dists = []
        for t in range(self.n_t):
            self.dists += [base_dist(hpar)]
        self.tau = [self.dists[t].tau for t in range(self.n_t)]
        self.hpar = [self.dists[t].hpar for t in range(self.n_t)]
        
    def hpar_table(self):
        '''
        Put all of the (conventional) hyperparameters in a table (Pandas data frame).
        '''
        hpar_names = list(self.dists[0].hpar.keys())
        table = pd.DataFrame(0.0, index = range(self.n_t), columns = hpar_names)
        for t in range(self.n_t):
            for hpar_name in hpar_names:
                table.loc[t, hpar_name] = self.dists[t].hpar[hpar_name]
        return table
        
    def update(self, y, z):
        for t in range(self.n_t):
            self.dists[t].update(y, weights = z[t,:])
            self.tau[t] = self.dists[t].tau
            self.hpar[t] = self.dists[t].hpar
            
    def log_density(self, theta):
        return np.sum([self.dists[t].log_density(theta[t]) for t in range(self.n_t)])
    
    def density(self, theta):
        return np.exp(self.log_density(theta))
    
    def log_marginal(self, y, z):
        z_flat = z.argmax(axis = 0) # encodes group in a single vector (not 1-hot encoding)
        return np.sum([self.dists[t].log_marginal(y[z_flat == t]) for t in range(self.n_t)])
    
    def frac_log_marginal(self, y, z, frac):
        return np.sum([self.dists[t].weighted_log_marginal(y, weights = frac[t]*z[t,:]) for t in range(self.n_t)])
    
    def marginal(self, y, z, frac):
        return np.exp(self.log_marginal(y, z, frac))
    
class distribution_with_predictor:
    '''
    
    Notes
    -----
    It is assumed that the last axis in T, z, S(z) etc. represents individual observations, e.g. participants.
    Thus, np.sum(T[j], axis = -1) takes the sum across observations.
    '''
        
    def tau_prime(self, y, z):
        T = self.T(y, z)
        n = y.shape[0]
        return [self.tau[j] + np.sum(T[j], axis = -1) for j in range(self.p)] + [self.tau[self.p] + n, self.tau[self.p + 1] + np.sum(self.S(z), axis = -1)]
    
    def weighted_tau_prime(self, y, z, weights):
        '''
        Notes
        -----
        weights represents a weighting factor (should be from 0 to 1) for each 
        observation.
        '''
        T = self.T(y, z)
        n = np.sum(weights)
        return [self.tau[j] + np.sum(weights*T[j], axis = -1) for j in range(self.p)] + [self.tau[self.p] + n, self.tau[self.p + 1] + np.sum(weights*self.S(z), axis = -1)] 
    
    def update(self, y, z):
        '''
        Perform conjugate prior updating.
        '''
        self.tau = self.tau_prime(y, z)
        self.hpar = self.tau_to_hpar(self.tau)
    
    def log_density(self, theta):
        '''
        Compute the log density of theta.
        '''
        return np.sum([np.sum(self.tau[j]*theta[j]) for j in range(self.p)]) - self.tau[self.p]*self.g(theta) - np.sum(self.tau[self.p + 1]*self.k(theta)) - self.h(self.tau)
        
    def density(self, theta):
        return np.exp(self.log_density(theta))
        
    #def log_predictive(self, y, z):
    #    '''
    #    Compute the log predictive distribution of data.
    #    '''
        # ADD THIS
        
    #def predictive(self, y, z):
    #    return np.exp(self.log_predictive(y, z))
    
    def log_marginal(self, y, z):
        '''
        Compute the log of the marginal distribution of data.
        '''
        return -np.sum(self.f(y)) - self.h(self.tau) + self.h(self.tau_prime(y, z))
    
    def weighted_log_marginal(self, y, z, weights):
        return -np.sum(weights*self.f(y)) - self.h(self.tau) + self.h(self.weighted_tau_prime(y, z, weights))
    
    def marginal(self, y, z):
        return np.exp(self.log_marginal(y, z))   
    
class beta_bernoulli(distribution):
    
    def __init__(self, hpar = {'a': 1.0, 'b': 1.0}):
        '''
        Initialize the beta distribution over theta (log-odds) with given parameters.
        '''
        self.hpar = hpar # conventional hyperparameters
        self.tau = self.hpar_to_tau(self.hpar)
        self.p = 1 # number of likelihood parameters
    
    def hpar_to_tau(self, hpar):
        '''
        Convert conventional hyperparameters to natural ones (tau).
        '''
        return [hpar['a'] - 1, hpar['a'] + hpar['b'] - 2]
    
    def tau_to_hpar(self, tau):
        '''
        Convert natural hyperparameters (tau) to conventional ones.
        '''
        return {'a': tau[1-1] + 1, 'b': tau[2-1] - tau[1-1] + 1}
    
    def T(self, y):
        '''
        Compute sufficient statistics.
        '''
        return [y]
        
    def f(self, y):
        return 0.0*y
        
    def g(self, theta):
        return np.log(1 + np.exp(theta))
        
    def h(self, tau):
        hpar = self.tau_to_hpar(tau)
        return betaln(hpar['a'], hpar['b'])

class normal_known_precision(distribution):
    '''
    Normal likelihood with known precision (xi) and a normal
    prior on the mean (mu).
    
    DOUBLE CHECK THIS (SOMETHING SEEMS TO BE WRONG).
    '''
    def __init__(self, hpar = {'m_mu': 0.0, 'xi_mu': 1.0}, xi = 1.0):
        self.xi = xi
        self.sigma = 1/np.sqrt(self.xi)
        self.hpar = hpar
        self.tau = self.hpar_to_tau(self.hpar)
        self.p = 1
        
    def hpar_to_tau(self, hpar):
        '''
        Convert conventional hyperparameters to natural ones (tau).
        '''
        return [self.hpar['m_mu']*self.sigma*self.hpar['xi_mu'], self.hpar['xi_mu']/self.xi]
    
    def tau_to_hpar(self, tau):
        '''
        Convert natural hyperparameters (tau) to conventional ones.
        '''
        return {'m_mu': self.sigma*self.tau[1-1]/self.tau[2-1], 'xi_mu': self.xi*self.tau[2-1]}
    
    def par_to_theta(self, par):
        '''
        Convert conventional parameters to natural ones (theta).
        '''
        return np.array(par['mu']/self.sigma)
    
    def T(self, y):
        return [y/self.sigma]
    
    def f(self, y):
        return 0.5*(self.xi*y**2 + np.log(2*np.pi) - np.log(self.xi))
    
    def g(self, theta):
        return 0.5*theta**2
        
    def h(self, tau):
        hpar = self.tau_to_hpar(tau)
        return 0.5*(np.log(2*np.pi) + hpar['xi_mu']*hpar['m_mu']**2 - np.log(hpar['xi_mu']) + np.log(self.xi))
    
class normal_gamma_normal(distribution):
    
    def __init__(self, hpar = {'m': 0.0, 'lambda': 1.0, 'alpha': 1.0, 'beta': 1.0}):
        self.hpar = hpar
        self.tau = self.hpar_to_tau(self.hpar)
        self.p = 2
    
    def hpar_to_tau(self, hpar):
        '''
        Convert conventional hyperparameters to natural ones (tau).
        '''
        return [hpar['lambda']*hpar['m'],
                -hpar['beta'] - 0.5*hpar['lambda']*hpar['m']**2,
                2*hpar['alpha'] - 1]
    
    def tau_to_hpar(self, tau):
        '''
        Convert natural hyperparameters (tau) to conventional ones.
        '''
        return {'m': tau[1-1]/tau[3-1],
                'lambda': tau[3-1],
                'alpha': 0.5*(tau[3-1] + 1),
                'beta': -0.5*tau[1-1]**2/tau[3-1] - tau[2-1]}
    
    def par_to_theta(self, par):
        '''
        Convert conventional parameters to natural ones (theta).
        '''
        return np.array([par['mu']/par['sigma']**2, 1/par['sigma']**2])
    
    def T(self, y):
        return [y, -0.5*y**2]      
        
    def f(self, y):
        n = y.shape[0]
        return np.array(n*[0.5*np.log(2*np.pi)])
        
    def g(self, theta):
        return 0.5*(theta[1-1]**2/theta[2-1] - np.log(theta[2-1]))
        
    def h(self, tau):
        hpar = self.tau_to_hpar(tau)
        return 0.5*np.log(2*np.pi) + gammaln(hpar['alpha']) - hpar['alpha']*np.log(hpar['beta']) - 0.5*np.log(hpar['lambda'])
    
class multi_group_normal(distribution_with_predictor):
    '''
    Notes
    -----
    z is an n_t x n array indicating group membership.
    z[t, i] = 1 if observation i is in group t and = 0 otherwise.
    
    theta = [xi*mu, xi]
    mu = the vector of the means for y in each group
    xi = the precision of y (shared across groups)
    '''
    
    def __init__(self, hpar = {'m': 0.0, 'lambda': 1.0, 'alpha': 1.0, 'beta': 1.0}, n_t = 2):
        '''
        Parameters
        ----------
        hpar: dict
            Initial (conventional) hyperparameters.
        n_t: int
            Number of groups.
        '''
        self.n_t = n_t
        self.hpar = {'m': hpar['m']*np.ones(n_t), 
                     'lambda': hpar['lambda']*np.ones(n_t), 
                     'alpha': hpar['alpha'], 
                     'beta': hpar['alpha']}
        self.tau = self.hpar_to_tau(self.hpar)
        self.p = 2
    
    def hpar_to_tau(self, hpar):
        '''
        Convert conventional hyperparameters to natural ones (tau).
        '''
        return [hpar['lambda']*hpar['m'], 
                -hpar['beta'] - 0.5*np.sum(hpar['lambda']*hpar['m']**2),
                2*hpar['alpha'] + self.n_t - 2,
                0.5*hpar['lambda']]
    
    def tau_to_hpar(self, tau):
        return {'m': 0.5*tau[1-1]/tau[4-1],
                'lambda': 2*tau[4-1],
                'alpha': 0.5*(tau[3-1] - self.n_t) + 1,
                'beta': -tau[2-1] - 0.25*np.sum(tau[1-1]**2/tau[4-1])}
    
    def T(self, y, z):
        return [z*y, -0.5*y**2]
    
    def S(self, z):
        return 0.5*z
    
    def f(self, y):
        n = y.shape[0]
        return np.array(n*[0.5*np.log(2*np.pi)])
        
    def g(self, theta):
        return -0.5*np.log(theta[2-1])
        
    def k(self, theta):
        return theta[1-1]**2/theta[2-1]
        
    def h(self, tau):
        hpar = self.tau_to_hpar(tau)
        return 0.5*self.n_t*np.log(2*np.pi) + gammaln(hpar['alpha']) - hpar['alpha']*np.log(hpar['beta']) - 0.5*np.sum(np.log(hpar['lambda']))
        
    


import numpy as np
from . import exp_families

def normal_shared_var(y, z, prior_m = 0.0, prior_lambda = 1.0, prior_alpha = 1.0, prior_beta = 1.0):
    (y, z) = _remove_nan(y, z)
    n_t = z.shape[0] # number of groups
    hpar = {'m': prior_m, 'lambda': prior_lambda, 'alpha': prior_alpha, 'beta': prior_beta} # prior hyperparameters
    H1 = exp_families.multi_group_normal(hpar, n_t)
    H0 = exp_families.normal_gamma_normal(hpar)
    log_bf10 = H1.log_marginal(y, z) - H0.log_marginal(y)
    return np.exp(log_bf10)

def post_normal_shared_var(y, z, prior_m = 0.0, prior_lambda = 1.0, prior_alpha = 1.0, prior_beta = 1.0):
    (y, z) = _remove_nan(y, z)
    n_t = z.shape[0] # number of groups
    hpar = {'m': prior_m, 'lambda': prior_lambda, 'alpha': prior_alpha, 'beta': prior_beta} # prior hyperparameters
    H1 = exp_families.multi_group_normal(hpar, n_t)
    H0 = exp_families.normal_gamma_normal(hpar)
    H1.update(y, z)
    H0.update(y)
    log_bf10 = H1.log_marginal(y, z) - H0.log_marginal(y)
    return np.exp(log_bf10)

def frac_normal_shared_var(y, z, prior_m = 0.0, prior_lambda = 1.0, prior_alpha = 1.0, prior_beta = 1.0):
    (y, z) = _remove_nan(y, z)    
    n_per_group = z.sum(axis = -1)
    z_flat = z.argmax(axis = 0) # encodes group in a single vector (not 1-hot encoding)
    frac = 1/np.sqrt(n_per_group) # different training fractions for each group
    n_t = z.shape[0] # number of groups
    hpar = {'m': prior_m, 'lambda': prior_lambda, 'alpha': prior_alpha, 'beta': prior_beta} # prior hyperparameters
    H1 = exp_families.multi_group_normal(hpar, n_t)
    H0 = exp_families.normal_gamma_normal(hpar)
    log_bf10 = (H1.log_marginal(y, z) - H1.weighted_log_marginal(y, z, weights = frac[z_flat])) - (H0.log_marginal(y) - H0.weighted_log_marginal(y, weights = frac[z_flat]))
    return np.exp(log_bf10)

def normal_dif_var(y, z, prior_m = 0.0, prior_lambda = 1.0, prior_alpha = 1.0, prior_beta = 1.0):
    (y, z) = _remove_nan(y, z)
    n_t = z.shape[0] # number of groups
    hpar = {'m': prior_m, 'lambda': prior_lambda, 'alpha': prior_alpha, 'beta': prior_beta} # prior hyperparameters
    H1 = exp_families.multi_group_distribution(exp_families.normal_gamma_normal, hpar, n_t)
    H0 = exp_families.normal_gamma_normal(hpar)
    log_bf10 = H1.log_marginal(y, z) - H0.log_marginal(y)
    return np.exp(log_bf10)

def frac_normal_dif_var(y, z, prior_m = 0.0, prior_lambda = 1.0, prior_alpha = 1.0, prior_beta = 1.0):
    (y, z) = _remove_nan(y, z)
    n_per_group = z.sum(axis = -1)
    z_flat = z.argmax(axis = 0) # encodes group in a single vector (not 1-hot encoding)
    frac = 1/np.sqrt(n_per_group) # different training fractions for each group
    n_t = z.shape[0] # number of groups
    hpar = {'m': prior_m, 'lambda': prior_lambda, 'alpha': prior_alpha, 'beta': prior_beta} # prior hyperparameters
    H1 = exp_families.multi_group_distribution(exp_families.normal_gamma_normal, hpar, n_t)
    H0 = exp_families.normal_gamma_normal(hpar)
    log_bf10 = (H1.log_marginal(y, z) - H1.frac_log_marginal(y, z, frac = frac)) - (H0.log_marginal(y) - H0.weighted_log_marginal(y, weights = frac[z_flat]))
    return np.exp(log_bf10)

def bernoulli(y, z, prior_a = 0.5, prior_b = 0.5):
    (y, z) = _remove_nan(y, z)
    n_t = z.shape[0] # number of groups
    H1 = exp_families.multi_group_distribution(exp_families.beta_bernoulli, 
                                               {'a': prior_a, 'b': prior_b}, 
                                               n_t)
    H0 = exp_families.beta_bernoulli({'a': prior_a, 'b': prior_b})
    log_bf10 = H1.log_marginal(y, z) - H0.log_marginal(y)
    return np.exp(log_bf10)

def post_bernoulli(y, z, prior_a = 0.5, prior_b = 0.5):
    (y, z) = _remove_nan(y, z)
    n_t = z.shape[0] # number of groups
    H1 = exp_families.multi_group_distribution(exp_families.beta_bernoulli, 
                                               {'a': prior_a, 'b': prior_b}, 
                                               n_t)
    H0 = exp_families.beta_bernoulli({'a': prior_a, 'b': prior_b})
    H1.update(y, z)
    H0.update(y)
    log_bf10 = H1.log_marginal(y, z) - H0.log_marginal(y)
    return np.exp(log_bf10)

def frac_bernoulli(y, z, prior_a = 0.5, prior_b = 0.5):
    (y, z) = _remove_nan(y, z)
    n_per_group = z.sum(axis = -1)
    z_flat = z.argmax(axis = 0) # encodes group in a single vector (not 1-hot encoding)
    frac = 1/np.sqrt(n_per_group) # different training fractions for each group
    n_t = z.shape[0] # number of groups
    H1 = exp_families.multi_group_distribution(exp_families.beta_bernoulli, 
                                               {'a': prior_a, 'b': prior_b}, 
                                               n_t)
    H0 = exp_families.beta_bernoulli({'a': prior_a, 'b': prior_b})
    log_bf10 = (H1.log_marginal(y, z) - H1.frac_log_marginal(y, z, frac = frac)) - (H0.log_marginal(y) - H0.weighted_log_marginal(y, weights = frac[z_flat]))
    return np.exp(log_bf10)

def _remove_nan(y, z):
    '''
    Convenience function to remove observations with missing data.
    '''
    is_nan = np.isnan(y)
    if np.sum(is_nan) > 0:
        y = y[~is_nan].copy()
        z = z[:, ~is_nan].copy()
    return (y, z)


import numpy as np
import pandas as pd
import plotnine as p9
from copy import copy
from itertools import combinations, product
from scipy.stats import norm as norm_dist
from scipy.stats import beta as beta_dist
from scipy.stats import multinomial as multinomial_dist
from scipy.stats import gamma as gamma_dist
from scipy.stats import t as t_dist
from scipy.special import digamma, loggamma
from scipy.spatial.distance import pdist, squareform

class lpa_model:
    '''
    Bayesian latent profile analysis (LPA) model with a Dirichlet process prior to
    allow for a potentially unbounded number of latent profiles.
    
    Latent profiles have different means for each variable but share a common precision (xi).
    These have a normal-gamma prior.
    
    The concentration parameter (alpha) is learned.
    
    Attributes
    ----------
    #####
    
    Methods
    -------
    
    Notes
    -----
    This is based on Blei and Jordan (2006).
    
    The concentration parameter is given a Gamma(3, 1) prior and learned via variational
    Bayes as described in Appendix B.
    '''
    
    def __init__(self, x, T = 20, prior_w1 = 3.0, prior_w2 = 1.0, prior_lambda_mu = 0.5, prior_strength_for_xi = 10.0, seed = 1234):
        """
        Parameters
        ----------
        x: data frame or array
            Observed data (data frame or matrix).  If a matrix then rows are variables and columns
            are observations; if a data frame then rows are observations and columns are variables.
        T: integer, optional
            Truncation level for the variational approximation.
        prior_w1: float, optional,
            First prior hyperparameter on alpha.
        prior_w2: float, optional
            Second prior hyperparameter on alpha.
        prior_lambda_mu: float, optional
            Controls the width of the prior distribution on mu: 
            higher values -> tighter prior around 0.
        prior_strength_for_xi: float, optional
            Controls the strength of the gamma prior distribution on xi.
            This prior is assumed to have a mean of 1, with alpha = prior_strength_for_xi/2
            and beta = prior_strength_for_xi/2.
        seed: integer, optional
            The seed for random number generation.
            
        Notes
        -----
        This is based on Blei and Jordan (2006).

        The concentration parameter (alpha) is given a Gamma(prior_w1, prior_w2) prior and learned via variational
        Bayes as described in Appendix B.
        
        Thus the prior mean of alpha is prior_w1/prior_w2.

        The means (mu) and precisions (xi) have a multi-mean normal-gamma prior.
        
        The fact that the prior alpha_xi = prior beta_xi implies that the prior distribution on xi
        has a mean of 1.
        """
        # ----- SET UP SOME VARIABLES -----
        
        if isinstance(x, pd.DataFrame):
            self.x = x.values.copy().transpose()
            self.index = x.index.values
            self.x_names = pd.Categorical(x.columns.values, categories = x.columns.values, ordered = True)
        else:
            self.x = x
            self.index = None
            x_names = ['var ' + str(i) for i in range(self.x.shape[0])] # default variable names
            self.x_names = pd.Categorical(x_names, categories = x_names, ordered = True)
        self.mask = 1 - np.isnan(self.x) # = 1 if not missing, = 0 if missing
        self.x[np.isnan(self.x)] = 0 # this is an arbitrary value to make computations work
        self.n = self.x.shape[1] # number of observations
        self.n_per_var = self.mask.sum(axis = 1) # number of observations for each variable (accounting for missing data)
        self.m = self.x.shape[0] # number of observed variables
        self.T = T
        self.seed = seed
        self.rng = np.random.default_rng(seed = self.seed) # for generating random permutations
        # lists for storing the ELBO and its components at each iteration
        self.E_log_lik_list = []
        self.E_log_prior_list = []
        self.entropy_list = []
        self.elbo_list = []
        # prior conventional hyperparameters
        self.prior_alpha_xi = 0.5*prior_strength_for_xi
        self.prior_beta_xi = 0.5*prior_strength_for_xi
        self.prior_m_mu = 0.0
        self.prior_lambda_mu = prior_lambda_mu
        # prior natural hyperparameters
        self.prior_tau1 = self.prior_lambda_mu*self.prior_m_mu
        self.prior_tau2 = -self.prior_beta_xi - 0.5*self.T*self.prior_lambda_mu*self.prior_m_mu**2
        self.prior_tau3 = 2*self.prior_alpha_xi + self.T - 2
        self.prior_tau4 = 0.5*self.prior_lambda_mu
        self.h_prior_tau = 0.5*self.T*np.log(2*np.pi) + loggamma(self.prior_alpha_xi) - self.prior_alpha_xi*np.log(self.prior_beta_xi) - 0.5*self.T*np.log(self.prior_lambda_mu) # prior log-normalizer for each of the observed variables

        # ----- INITIALIZE VARIATIONAL HYPERPARAMETERS -----
        
        # initialize z's hyperparameters
        self.phi = np.zeros([self.T, self.n])
        for i in range(self.n):
            self.phi[:,i] = self.rng.dirichlet(self.T*[1.0])
        # initialize hyperparameters for mu and xi
        self.tau2 = self.prior_tau2 - 0.5*np.sum(self.mask*self.x**2, axis = 1)
        self.tau3 = self.prior_tau3 + self.n_per_var
        self.tau1 = np.zeros([self.m, self.T])
        self.tau4 = np.zeros([self.m, self.T])
        for j in range(self.m):
            for t in range(self.T):
                self.tau1[j,t] = self.prior_tau1 + np.sum(self.mask[j,:]*self.phi[t,:]*self.x[j,:])
                self.tau4[j,t] = self.prior_tau4 + 0.5*np.sum(self.mask[j,:]*self.phi[t,:])
        self.update_conventional_hyperparameters()
        # compute relevant expected values etc.
        self.v_mu = np.zeros([self.m, self.T])
        self.E_xi_mu = np.zeros([self.m, self.T])
        self.E_xi_mu2 = np.zeros([self.m, self.T])
        self.update_expectations_etc()
        # initialize alpha's hyperparameters
        self.prior_w1 = prior_w1
        self.prior_w2 = prior_w2
        self.w1 = self.prior_w1 + self.T - 1
        self.w2 = self.prior_w2
        self.E_alpha = self.w1/self.w2
        self.E_log_alpha = digamma(self.w1) - np.log(self.w2)
        # initialize V's hyperparameters
        self.gamma1 = np.zeros(self.T)
        self.gamma2 = np.zeros(self.T)
        self.E_log_V = np.zeros(self.T)
        self.E_log_1minusV = np.zeros(self.T)
        for t in range(self.T):
            self.gamma1[t] = 1 + np.sum(self.phi[t,:])
            self.gamma2[t] = self.E_alpha + np.sum(self.phi[range(t+1,self.T),:])
            self.E_log_V[t] = digamma(self.gamma1[t]) - digamma(self.gamma1[t] + self.gamma2[t])
            self.E_log_1minusV[t] = digamma(self.gamma2[t]) - digamma(self.gamma1[t] + self.gamma2[t])
        
        # ----- OTHER STUFF -----
        
        self.E_pi = self.phi.mean(axis = 1)
        self._update_hard_assignment()
    
    def fit(self, tolerance = 1e-05, max_iter = 50, min_iter = 20):
        """
        Parameters
        ----------
        tolerance: float, optional
            Relative change in the ELBO at which the optimization should stop.
        max_iter: integer, optional
            Maximum number of iterations to run the optimization.
        min_iter: integer, optional
            Minimum number of iterations to run the optimization.
        """
        
        relative_elbo_change = 1.0 # this is an arbitrary value for the first iteration
        itr = 0 # counter for iterations
        continue_loop = True
        int_elbo = -np.Inf        
        
        while continue_loop:
            # ----- UPDATE VARIATIONAL HYPERPARAMETERS -----
            
            # update z's hyperparameters
            for i in range(self.n):
                S = np.zeros(self.T)
                for t in range(self.T):
                    log_lik_term = np.sum(self.mask[:,i]*(self.E_xi_mu[:,t]*self.x[:,i] - 0.5*self.E_xi*self.x[:,i]**2 - 0.5*np.log(2*np.pi) - 0.5*self.E_xi_mu2[:,t] + 0.5*self.E_log_xi))
                    S[t] = self.E_log_V[t] + np.sum(self.E_log_1minusV[range(t-1)]) + log_lik_term
                S_max = np.max(S)
                log_sum_exp_S = S_max + np.log(np.sum(np.exp(S - S_max))) # = log(sum(exp(S))) but uses log-sum-exp trick for numerical accuracy
                log_phi = S - log_sum_exp_S
                self.phi[:,i] = np.exp(log_phi)
                self.phi[:,i] /= self.phi[:,i].sum() # shouldn't need to renormalize, but it helps with some rounding errors in annoying cases
                
            # update V's hyperparameters
            for t in range(self.T):
                self.gamma1[t] = 1 + np.sum(self.phi[t,:])
                self.gamma2[t] = self.E_alpha + np.sum(self.phi[range(t+1,self.T),:])
                self.E_log_V[t] = digamma(self.gamma1[t]) - digamma(self.gamma1[t] + self.gamma2[t])
                self.E_log_1minusV[t] = digamma(self.gamma2[t]) - digamma(self.gamma1[t] + self.gamma2[t])         
            
            # update hyperparameters for mu and xi
            for j in range(self.m):
                # tau2 and tau3 don't change, so we don't update them
                for t in range(self.T):
                    self.tau1[j,t] = self.prior_tau1 + np.sum(self.mask[j,:]*self.phi[t,:]*self.x[j,:])
                    self.tau4[j,t] = self.prior_tau4 + 0.5*np.sum(self.mask[j,:]*self.phi[t,:])
            self.update_conventional_hyperparameters()        
            self.update_expectations_etc() # compute relevant expected values etc.          
            
            # update alpha's hyperparameters
            self.w2 = self.prior_w2 - np.sum(self.E_log_1minusV[range(self.T - 1)]) # self.w1 stays constant, so we only update self.w2
            self.E_alpha = self.w1/self.w2
            self.E_log_alpha = digamma(self.w1) - np.log(self.w2)

            # ----- COMPUTE ELBO -----
            
            self.E_log_lik_list += [self.compute_E_log_lik()]
            self.E_log_prior_list += [self.compute_E_log_prior()]
            self.entropy_list += [self.compute_entropy()]
            self.elbo_list += [self.E_log_lik_list[-1] + self.E_log_prior_list[-1] + self.entropy_list[-1]]
            
            # ----- FINISH CURRENT ITERATION -----      
            
            itr += 1 # increment the iteration counter
            if itr < min_iter: # automatically continue if below the minimum number of iterations
                continue_loop = True
            else: # otherwise, continue if ELBO change is below threshold and we haven't yet hit the iteration limit
                relative_elbo_change = np.abs((self.elbo_list[-2] - self.elbo_list[-1])/self.elbo_list[-2])
                continue_loop = (relative_elbo_change > tolerance) and (itr < max_iter)
                
        # ----- FINISH UP -----
        self.E_V = self.gamma1/(self.gamma1 + self.gamma2)
        product_term = np.concatenate([np.array([1]), np.cumprod(1 - self.E_V)[range(self.T-1)]])
        self.E_pi = self.E_V*product_term
        self._update_hard_assignment()
    
    def _update_hard_assignment(self):
        # figure out which profiles are non-empty and sort them by count
        unsorted_z_hat = self.phi.argmax(axis = 0) # unsorted point estimates of z (hard assignment)
        unsorted_z_hat_1hot = np.zeros(self.phi.shape)
        for t in range(self.T):
            unsorted_z_hat_1hot[t, unsorted_z_hat == t] = 1.0
        unsorted_counts = unsorted_z_hat_1hot.sum(axis = -1)
        self.n_profiles = np.sum(unsorted_counts > 0) # number of non-empty profiles
        self.profile_order = np.argsort(unsorted_counts)[::-1][range(self.n_profiles)]
        
        # create sorted versions of phi, m_mu etc. that only include non-empty profiles
        self.sorted_phi = self.phi[self.profile_order, :]
        self.sorted_m_mu = self.m_mu[:, self.profile_order]
        self.sorted_v_mu = self.v_mu[:, self.profile_order]
        
        # sorted point estimates of z
        self.z_hat = self.sorted_phi.argmax(axis = 0) # point estimates of z (hard assignment)
        self.z_hat_1hot = np.zeros(self.sorted_phi.shape) # 1-hot matrix of profile assignment
        for t in range(self.n_profiles):
            self.z_hat_1hot[t, self.z_hat == t] = 1.0
        self.n_per_profile = self.z_hat_1hot.sum(axis = -1) # count how many people are in each profile
        self.prop_per_profile = self.n_per_profile/self.n # proportion of people in each profile
    
    def update_conventional_hyperparameters(self):
        # convert natural hyperparameters to conventional ones
        self.alpha_xi = 0.5*(self.tau3 - self.T) - 1
        self.beta_xi = -self.tau2 - 0.25*np.sum((self.tau1**2)/self.tau4, axis = 1)
        self.m_mu = 0.5*self.tau1/self.tau4
        self.lambda_mu = 2*self.tau4
    
    def update_expectations_etc(self):
        self.E_xi = self.alpha_xi/self.beta_xi
        self.E_log_xi = digamma(self.alpha_xi) - np.log(self.beta_xi)
        self.h_tau = 0.5*self.T*np.log(2*np.pi) + loggamma(self.alpha_xi) - self.alpha_xi*np.log(self.beta_xi) - 0.5*np.sum(np.log(self.lambda_mu)) # log-normalizer
        for t in range(self.T):
            self.v_mu[:,t] = self.beta_xi/(self.lambda_mu[:,t]*(self.alpha_xi - 1))
            self.E_xi_mu[:,t] = (self.alpha_xi/self.beta_xi)*self.m_mu[:,t]
            self.E_xi_mu2[:,t] = (self.alpha_xi/self.beta_xi)*self.m_mu[:,t]**2 + 1/self.lambda_mu[:,t]
    
    def compute_E_log_lik(self):
        # compute the variational expectation of the total log-likelihood (can make more efficient later)
        E_log_lik = - 0.5**np.log(2*np.pi) + 0.5*np.sum(self.n_per_var*self.E_log_xi)
        for j in range(self.m):
            E_log_lik += np.sum( self.mask[j,:]*( np.inner(self.E_xi_mu[j,:], self.phi.T)*self.x[j,:] - 0.5*self.E_xi[j]*self.x[j,:]**2 - 0.5*np.inner(self.E_xi_mu2[j,:], self.phi.T) ) )
        return E_log_lik
    
    def compute_E_log_prior(self):
        # compute expected log-prior
        E_log_mu_xi_prior = np.sum(np.sum(self.prior_tau1*self.E_xi_mu, axis = 1) + self.prior_tau2*self.E_xi + 0.5*self.prior_tau3*self.E_log_xi - np.sum(self.prior_tau4*self.E_xi_mu2, axis = 1) - self.h_prior_tau)
        E_log_z_prior = 0.0
        for t in range(self.T): # this is taken from page 129 of Blei and Jordan (2006)
            q_z_greater_than_t = np.sum(self.phi[range(t+1, self.T), :], axis = 0)
            E_log_z_prior += np.sum(q_z_greater_than_t*self.E_log_1minusV[t] + self.phi[t,:]*self.E_log_V[t])
        E_log_V_prior = np.sum((self.E_alpha - 1)*self.E_log_1minusV + self.E_log_alpha)
        E_log_alpha_prior = -self.w2*self.E_alpha + (self.w1 - 1)*self.E_log_alpha - loggamma(self.w1) + self.w1*np.log(self.w2)
        return E_log_mu_xi_prior + E_log_z_prior + E_log_V_prior + E_log_alpha_prior

    def compute_entropy(self):
        # compute entropy
        H_mu_xi = np.sum(-np.sum(self.tau1*self.E_xi_mu, axis = 1) - self.tau2*self.E_xi - 0.5*self.tau3*self.E_log_xi + np.sum(self.tau4*self.E_xi_mu2, axis = 1) + self.h_tau)
        H_z = 0.0
        for i in range(self.n):
            H_z += multinomial_dist.entropy(1, self.phi[:,i])
        H_V = beta_dist.entropy(self.gamma1, self.gamma2).sum()
        H_alpha = gamma_dist.entropy(a = self.w1, scale = 1/self.w2)
        return H_mu_xi + H_z + H_V + H_alpha
    
    def profile_similarity(self, figure_size = (6, 5), font_size = 11, start_profile_labels_from1 = False):
        '''
        Computes pairwise similarity between latent profiles.
        
        Parameters
        ----------
        
        Notes
        -----
        This computes the Euclidean distance between each pair of profile means
        (m_mu) and converts it to a similarity measure as follows:
        
        squared_distance = sum(E_xi*(m_mu[:,j] - m_mu[:,k])^2)
        similarity = exp(-0.5*squared_distance)
        
        Two profiles with identical means will thus have a similarity of 1,
        while highly dissimilar profiles will have a similarity near 0.
        
        Distances are scaled by the estimated precision (E_xi) of each data
        dimension, so that variables with lower variances (higher E_xi) are
        given more weight.
        
        This similarity metric is proportional to the likelihood from each
        latent profile of the most typical (modal = mean) data point corresponding
        to each other latent profile (it is a diagonal multivariate normal density
        without the normalizing term).
        
        Profiles are numbered according to profile_order in the output, i.e.
        the profile with the largest estimated membership/base rate (pi)
        will be 0, the profile with the next largest membership will be 1
        etc.
        '''
        # compute similarity
        squared_distance = pdist(self.sorted_m_mu.T, metric = 'seuclidean', V = 1/self.E_xi)**2 # distances, scaled by estimated precision (E_xi)
        similarity = np.exp(-0.5*squareform(squared_distance)) # similarity matrix
        similarity_off_diag = similarity - 999*np.eye(similarity.shape[0]) # - 999*np.eye(similarity.shape[0]) effectively excludes the diagonal of the similarity matrix, which represents self-similarity (always 1)
        most_similar_pair = np.unravel_index(np.argmax(similarity_off_diag), similarity.shape) # find the most similar pair of profiles
        similarity_vector = np.exp(-0.5*squared_distance) # similarity vector (for taking the mean without redundant values)
        
        # make a plot of the two most similar profiles
        plot_df = pd.DataFrame({'mu': np.concatenate([self.sorted_m_mu[:, most_similar_pair[0]], self.sorted_m_mu[:, most_similar_pair[1]]]), 
                                'profile': self.m*['profile ' + str(most_similar_pair[0] + start_profile_labels_from1)] + self.m*['profile ' + str(most_similar_pair[1] + start_profile_labels_from1)], 
                                'variable': pd.Categorical(2*list(self.x_names), categories = self.x_names, ordered = True)})
        plot = p9.ggplot(plot_df, p9.aes(x = 'variable', y = 'mu')) + p9.geom_bar(stat = 'identity')
        plot += p9.theme_classic(base_size = font_size) + p9.theme(figure_size = figure_size, axis_text_x = p9.element_text(rotation = 90, hjust = 1))
        plot += p9.facet_wrap('profile', scales = 'free_x', ncol = 2)
        
        return {'similarity': similarity, 'max_similarity': similarity_off_diag.max(), 'most_similar_pair': most_similar_pair, 'mean_similarity': np.mean(similarity_vector), 'plot': plot}            
    
    def classify_new_data(self, x_new):
        '''
        Compute the (approximate) probability of new data points belonging to
        each profile given the model's current variational parameters.
        
        Notes
        -----
        The returned estimates are NOT sorted according to profile membership counts (profile_order).
        '''
        # set up variables
        if isinstance(x_new, pd.DataFrame):
            x_new = x_new.values.transpose()
        mask_new = 1 - np.isnan(x_new) # = 1 if not missing, = 0 if missing
        x_new[np.isnan(x_new)] = 0 # this is an arbitrary value to make computations work
        n_new = x_new.shape[1] # number of observations
        n_per_var_new = mask_new.sum(axis = 1) # number of observations for each variable (accounting for missing data)
        
        # compute log base rates
        E_V = self.gamma1/(self.gamma1 + self.gamma2)
        log_base_rate = np.log(E_V) + np.concatenate([np.array([0]), np.cumsum(np.log(1 - E_V))[range(self.T-1)]])
        
        # compute log posterior predictive terms
        log_post_pred = np.zeros([self.m, self.T, n_new])
        for j in range(self.m):
            for t in range(self.T):
                # https://en.wikipedia.org/wiki/Conjugate_prior#When_likelihood_function_is_a_continuous_distribution ** DOUBLE CHECK THIS **
                t_dist_sd = np.sqrt(self.beta_xi[j]*(self.lambda_mu[j, t] + 1)/(self.alpha_xi[j]*self.lambda_mu[j, t]))
                log_post_pred[j, t, :] = t_dist.logpdf(x_new[j, :],
                                                       df = 2*self.alpha_xi[j],
                                                       loc = self.m_mu[j, t],
                                                       scale = t_dist_sd)
                log_post_pred[j, t, :] *= mask_new[j, :]
        
        # combine them
        phi_new = np.zeros([self.T, n_new])
        for i in range(n_new):
            S = np.zeros(self.T)
            for t in range(self.T):
                S[t] = log_base_rate[t] + np.sum(log_post_pred[:, t, i])
            S_max = np.max(S)
            log_sum_exp_S = S_max + np.log(np.sum(np.exp(S - S_max))) # = log(sum(exp(S))) but uses log-sum-exp trick for numerical accuracy
            log_phi_new = S - log_sum_exp_S
            phi_new[:, i] = np.exp(log_phi_new)/np.exp(log_phi_new).sum()
            
        return phi_new
    
    def print_summary(self, start_profile_labels_from1 = False):
        '''
        Display information about the model.
        '''
        print('n = ' + str(self.n))
        
        H_z = 0.0
        for i in range(self.n):
            H_z += multinomial_dist.entropy(1, self.phi[:,i]) # using the entropy function avoids problems with very small logs
        max_H_z = self.n*np.log(self.T)
        print('\nM-Plus style entropy statistic: ' + str(np.round(1 - H_z/max_H_z, 2)))
        print('(= 1 - entropy of z/max entropy of z)')
        # see http://www.statmodel.com/download/UnivariateEntropy.pdf
        
        print('\nnumber of non-empty profiles: ' + str(self.n_profiles))
        if start_profile_labels_from1:
            profile_names = ['profile ' + str(i + 1) for i in range(self.n_profiles)]
        else:
            profile_names = ['profile ' + str(i) for i in range(self.n_profiles)]
        print('\npi (profile probabilties): ')
        print(np.round(pd.DataFrame(self.prop_per_profile, index = profile_names, columns = ['']).T, 2))
        print('\nnumber in each profile: ')
        print(pd.DataFrame(self.n_per_profile.astype(int), index = profile_names, columns = ['']).T)
        print('\nalpha (concentration parameter): ' + str(np.round(self.E_alpha, 2)))
        print('\nmu: (means)')
        print(np.round(pd.DataFrame(self.sorted_m_mu, index = self.x_names, columns = profile_names), 2))
        print('\nxi: (precisions, i.e. inverse variances)')
        print(np.round(pd.DataFrame(self.E_xi, index = self.x_names, columns = ['']).T, 2))
        
        print('\nproportion of the sample contained in the K biggest profiles')
        print(np.round(pd.DataFrame(self.prop_per_profile.cumsum(),
                                    index = ['K = ' + str(K) for K in np.arange(self.n_profiles) + 1],
                                    columns = ['']).T,
                       2))
    
    def print_univariate_entropy_stats(self):
        '''
        Compute and print M-Plus style univariate entropy statistics.
        '''
        print('\nM-Plus style univariate entropy statistics')
        print('(= 1 - entropy of z given x[j,:]/max entropy of z)')
        univariate_H_z = np.zeros(self.m)
        for j in range(self.m):
            for i in range(self.n):
                S = np.zeros(self.T)
                for t in range(self.T):
                    log_lik_term = self.mask[j,i]*self.E_xi_mu[j,t]*self.x[j,i] - 0.5*self.E_xi[j]*self.x[j,i]**2 - 0.5*np.log(2*np.pi) - 0.5*self.E_xi_mu2[j,t] + 0.5*self.E_log_xi[j]
                    S[t] = self.E_log_V[t] + np.sum(self.E_log_1minusV[range(t-1)]) + log_lik_term
                S_max = np.max(S)
                log_sum_exp_S = S_max + np.log(np.sum(np.exp(S - S_max)))
                log_phi_given_xj = S - log_sum_exp_S
                phi_given_xj = np.exp(log_phi_given_xj)
                phi_given_xj /= phi_given_xj.sum()
                univariate_H_z[j] += multinomial_dist.entropy(1, phi_given_xj)
        max_H_z = self.n*np.log(self.T)
        univariate_entropy_stat = 1 - univariate_H_z/max_H_z
        print(np.round(pd.DataFrame(univariate_entropy_stat, index = self.x_names, columns = ['']).T, 2))
    
    def compute_x_hat(self):
        '''
        Compute posterior mean predicted observations (x_hat).
        '''
        return np.inner(self.m_mu, self.phi.T)
    
    def compute_residuals(self):
        '''
        Compute and residuals (actual observations - posterior mean predicted observations).
        
        Notes
        -----
        Technically the residuals are random variables (because we're doing Bayesian stats)
        and we are computing their means.
        '''
        x_hat = self.compute_x_hat() # predicted observations
        resid = self.x - x_hat
        return pd.DataFrame(resid.T, columns = self.x_names)
    
    def plot_residuals(self, figure_size = (6, 5), font_size = 11, ncol = 4, bins = 20):
        '''
        Plot residuals (actual observations - posterior mean predicted observations).
        
        Notes
        -----
        Technically the residuals are random variables (because we're doing Bayesian stats)
        and we are plotting their means.
        
        This plots a density histogram of residuals for each variable, plus the corresponding
        normal distribution (same mean and variance as the residuals) for comparison (as a
        blue line).
        '''
        # create histograms of residuals
        rdf = self.compute_residuals().stack().to_frame().reset_index().rename(columns = {'level_0': 'i', 'level_1': 'variable', 0: 'residuals'})
        plot = p9.ggplot(rdf, p9.aes(x = 'residuals', y = p9.after_stat('density')))
        plot += p9.geom_histogram(bins = bins, fill = 'gray')
        plot += p9.facet_wrap('variable', scales = 'free', ncol = ncol)
        plot += p9.theme_classic(base_size = font_size) + p9.theme(figure_size = figure_size)
        
        # add normal density plots for comparison
        r_values = np.arange(-3, 3, step = 0.01)
        pdf_df_list = []

        for var in self.x_names:
            r = rdf.loc[rdf['variable'] == var, 'residuals']
            m = r.mean()
            s = r.std()
            norm_pdf = norm_dist.pdf(r_values, m, s)
            pdf_df_list += [pd.DataFrame({'variable': var, 'x': r_values, 'norm_pdf': norm_pdf})]
        pdf_df = pd.concat(pdf_df_list)
        plot += p9.geom_line(p9.aes(x = 'x', y = 'norm_pdf'), data = pdf_df, color = 'blue')
        
        return plot
    
    def plot_profile_probs(self, figure_size = (6, 5), start_profile_labels_from1 = False):
        '''
        Plot estimated profile probabilities (pi).
        '''
        plot_df = pd.DataFrame({'profile': pd.Series(range(self.n_profiles), dtype = 'string'),
                                'pi': self.prop_per_profile})
        if start_profile_labels_from1:
            plot_df['profile'] += 1
        plot = p9.ggplot(plot_df, p9.aes(x = 'profile', y = 'pi')) + p9.geom_bar(stat = 'identity')
        return plot
    
    def plot_profile_means(self, figure_size = (6, 5), font_size = 11, kind = 'bar', facet_var = 'profile', ncol = 4, fancy_labels = True, start_profile_labels_from1 = False):
        '''
        Plot profile means.
        '''
        # make a dataframe for plot data
        plot_df = pd.DataFrame(product(range(self.m), range(self.n_profiles)),
                               columns = ['j', 'profile'])
        plot_df['variable'] = self.x_names[plot_df['j'].values]
        plot_df['mu'] = 0.0
        plot_df['v'] = 0.0
        plot_df['mu_minus'] = 0.0
        plot_df['mu_plus'] = 0.0
        for r in range(plot_df.shape[0]):
            plot_df.loc[r, 'mu'] = self.sorted_m_mu[plot_df.iloc[r]['j'], plot_df.iloc[r]['profile']]
            plot_df.loc[r, 'v'] = self.sorted_v_mu[plot_df.iloc[r]['j'], plot_df.iloc[r]['profile']]
        plot_df['mu_minus'] = plot_df['mu'] - 1.96*np.sqrt(plot_df['v'])
        plot_df['mu_plus'] = plot_df['mu'] + 1.96*np.sqrt(plot_df['v'])
        if fancy_labels and facet_var == 'profile':
            # add information about pi (i.e. how many data points are in the profile) to labels
            pct_for_labels = np.round(100*self.prop_per_profile).astype(int) # pi, represented as percentages
            pct_info_list = [ ' (' + str(pct_for_labels[plot_df['profile'][r]]) + '%)' for r in range(plot_df.shape[0])]
            if start_profile_labels_from1:
                plot_df['profile'] += 1
            plot_df['profile'] = plot_df['profile'].astype('string')
            plot_df['profile'] = plot_df['profile'].str.cat(pct_info_list) # add pi information (as percentages)
            plot_df['profile'] = 'profile ' + plot_df['profile'] # add the word 'profile'
            plot_df['profile'] = pd.Categorical(plot_df['profile'].values, 
                                                categories = plot_df['profile'].values[range(self.n_profiles)], 
                                                ordered = True)
            
        else:
            if start_profile_labels_from1:
                plot_df['profile'] += 1
            plot_df['profile'] = plot_df['profile'].astype('string')
            plot_df['profile'] = pd.Categorical(plot_df['profile'].values, 
                                                categories = plot_df['profile'].values[range(self.n_profiles)], 
                                                ordered = True)
        
        # make the plot
        if facet_var == 'profile':
            plot = p9.ggplot(plot_df, p9.aes(x = 'variable', y = 'mu', ymin = 'mu_plus', ymax = 'mu_minus'))
        elif facet_var == 'variable':
            plot = p9.ggplot(plot_df, p9.aes(x = 'profile', y = 'mu', ymin = 'mu_plus', ymax = 'mu_minus'))
        
        if kind == 'bar':
            plot += p9.geom_bar(stat = 'identity')
        elif kind == 'point':
            plot += p9.geom_point()
            plot += p9.geom_hline(yintercept = 0, linetype = 'dashed')
        elif kind == 'point_and_errorbar':
            plot += p9.geom_point()
            plot += p9.geom_errorbar()
            plot += p9.geom_hline(yintercept = 0, linetype = 'dashed')
        plot += p9.theme_classic(base_size = font_size)
        plot += p9.theme(figure_size = figure_size)
        
        if facet_var == 'profile':
            plot += p9.facet_wrap('profile', scales = 'free_x', ncol = ncol)
            plot += p9.theme(axis_text_x = p9.element_text(rotation = 90, hjust = 1))
        elif facet_var == 'variable':
            plot += p9.facet_wrap('variable', scales = 'free_x', ncol = ncol)
        
        return plot


import numpy as np
import functools
import multiprocessing
from itertools import product
from .discrete_finite import ibp_model as discrete_model
from .sparse_cont_finite import ibp_model as cont_model

def fit_ibp(x, n_workers = 4, restarts = 4, alpha_values = [1.0, 2.0, 4.0, 8.0], omega_values = [0.1, 0.2, 0.4, 0.8], xi = 1, p = 4, include_intercept = True, use_cont_model = False, stop_threshold = 1e-05, max_iter = 50, min_iter = 30):
    """
    Parameters
    ----------
    x: array
        Matrix of observed data.  Rows are variables and columns are observations.
    n_workers: int, optional
        Number of workers in the pool for parallel processing; should be less than
        or equal to the number of available CPUs.
    restarts: int, optional
        Number of random restarts per set of alpha and omega values.
    alpha_values: array or list of floats, optional
        Parameter of the Indian buffet process.
    omega_values: array or list of floats, optional
        Prior precision of A.
    p: integer, optional
        Number of latent factors used in the finite approximation to the Indian buffet process.
    include_intercept: logical, optional
        If True (default) then an intercept term is included.  If False then it is not.
    use_cont_model: logical, optional
        If False (default) then use the discrete model.  If True then use the sparse continuous model.
    stop_threshold: float, optional
        Relative change in the ELBO at which the optimization should stop.
    max_iter: integer, optional
        Maximum number of iterations to run the optimization.
    min_iter: integer, optional
        Minimum number of iterations to run the optimization.
    """
    model_list = []
    final_elbo = []
    
    with multiprocessing.Pool(n_workers) as pool:
        if use_cont_model:
            fun = functools.partial(__cont_model_fit_wrapper__, x = x, xi = xi, p = p, include_intercept = include_intercept, stop_threshold = stop_threshold, max_iter = max_iter, min_iter = min_iter)
        else:
            fun = functools.partial(__discrete_model_fit_wrapper__, x = x, xi = xi, p = p, include_intercept = include_intercept, stop_threshold = stop_threshold, max_iter = max_iter, min_iter = min_iter)
        alpha_omega_values = n_workers*list(product(alpha_values, omega_values))
        
        results = pool.starmap(fun, alpha_omega_values)
        for result in results:
            model_list += [result]
            final_elbo += [result.elbo_list[-1]]
        
    final_elbo = np.array(final_elbo)
        
    return {'model_list': model_list, 'final_elbo': final_elbo, 'best_model': model_list[final_elbo.argmax()]}

def __discrete_model_fit_wrapper__(alpha, omega, x, xi, p, include_intercept, stop_threshold, max_iter, min_iter):
    """
    Defines a discrete IBP model and fits it to the data, returning the result.
    This is only defined in order to get the parallelization to work.
    """
    new_model = discrete_model(x = x, alpha = alpha, omega = omega, xi = xi, p = p, include_intercept = include_intercept)
    new_model.fit(stop_threshold = stop_threshold, max_iter = max_iter, min_iter = min_iter)
    return new_model
    
def __cont_model_fit_wrapper__(alpha, omega, x, xi, p, include_intercept, stop_threshold, max_iter, min_iter):
    """
    Defines a sparse continuous IBP model and fits it to the data, returning the result.
    This is only defined in order to get the parallelization to work.
    """
    new_model = cont_model(x = x, alpha = alpha, omega = omega, xi = xi, p = p, include_intercept = include_intercept)
    new_model.fit(stop_threshold = stop_threshold, max_iter = max_iter, min_iter = min_iter)
    return new_model


import numpy as np
import functools
import multiprocessing
from itertools import product
from .lpa import lpa_model
from scipy.special import digamma

def fit_lpa(x, starting_hpar = None, n_workers = 4, restarts = 4, T = 20, prior_w1 = 3.0, prior_w2 = 1.0, prior_lambda_mu = 0.5, prior_strength_for_xi = 10.0, seed = 1234, tolerance = 1e-05, max_iter = 50, min_iter = 30):
    """
    Parameters
    ----------
    x: data frame or array
        Observed data (data frame or matrix).  If a matrix then rows are variables and columns
        are observations; if a data frame then rows are observations and columns are variables.
    starting_hpar: tuple, optional
        If None (default) then the initial hyperparameters will be sampled as normal.
        If specified, then it should be in as a dict with keys (phi, w1, w2, gamma1, gamma2).
    n_workers: int, optional
        Number of workers in the pool for parallel processing; should be less than
        or equal to the number of available CPUs.
    restarts: int, optional
        Number of random restarts per set of alpha and omega values.
    T: integer, optional
        Truncation level of the variational approximation.
    prior_w1: float, optional
        First prior hyperparameter on alpha.
    prior_w2: float, optional
        Second prior hyperparameter on alpha.
    prior_lambda_mu: float, optional
        Controls the width of the prior distribution on mu: 
        higher values -> tighter prior around 0.
    prior_strength_for_xi: float, optional
        Controls the strength of the gamma prior distribution on xi.
        This prior is assumed to have a mean of 1, with alpha = prior_strength_for_xi/2
        and beta = prior_strength_for_xi/2.
    seed: int, optional
        Random seed (determines starting conditions of each optimization).
    tolerance: float, optional
        Relative change in the ELBO at which the optimization should stop.
    max_iter: integer, optional
        Maximum number of iterations to run the optimization.
    min_iter: integer, optional
        Minimum number of iterations to run the optimization.
        
    Notes
    -----
    ***** ADD THESE *****
    """
    model_list = []
    final_elbo = []
    
    with multiprocessing.Pool(n_workers) as pool:
        fun = functools.partial(__model_fit_wrapper__, x = x, starting_hpar = starting_hpar, T = T, prior_w1 = prior_w1, prior_w2 = prior_w2, prior_lambda_mu = prior_lambda_mu, prior_strength_for_xi = prior_strength_for_xi, tolerance = tolerance, max_iter = max_iter, min_iter = min_iter)
        
        # generate random seeds for each model based on the seed parameter
        seed_list = []
        rng = np.random.default_rng(seed)
        for i in range(restarts):
            seed_list += [rng.integers(low = 1000, high = 9999)]
        
        results = pool.map(fun, seed_list)
        for result in results:
            model_list += [result]
            final_elbo += [result.elbo_list[-1]]
        
    final_elbo = np.array(final_elbo)
        
    return {'final_elbo': final_elbo, 'model_list': model_list, 'best_model': model_list[final_elbo.argmax()]}

def __model_fit_wrapper__(seed, x, starting_hpar, T, prior_w1, prior_w2, prior_lambda_mu, prior_strength_for_xi, tolerance, max_iter, min_iter):
    """
    Defines an LPA model and fits it to the data, returning the result.
    This function is only defined in order to get the parallelization to work.
    """
    new_model = lpa_model(x = x, T = T, prior_w1 = prior_w1, prior_w2 = prior_w2, prior_lambda_mu = prior_lambda_mu, prior_strength_for_xi = prior_strength_for_xi, seed = seed)
    if not starting_hpar is None:
        new_model.phi = starting_hpar['phi']
        new_model.w1 = starting_hpar['w1']
        new_model.w2 = starting_hpar['w2']
        new_model.E_alpha = new_model.w1/new_model.w2
        new_model.E_log_alpha = digamma(new_model.w1) - np.log(new_model.w2)
        new_model.gamma1 = starting_hpar['gamma1']
        new_model.gamma2 = starting_hpar['gamma2']
        for t in range(new_model.T):
            new_model.E_log_V[t] = digamma(new_model.gamma1[t]) - digamma(new_model.gamma1[t] + new_model.gamma2[t])
            new_model.E_log_1minusV[t] = digamma(new_model.gamma2[t]) - digamma(new_model.gamma1[t] + new_model.gamma2[t])
        for j in range(new_model.m):
            for t in range(new_model.T):
                new_model.tau1[j,t] = new_model.prior_tau1 + np.sum(new_model.mask[j,:]*new_model.phi[t,:]*new_model.x[j,:])
                new_model.tau4[j,t] = new_model.prior_tau4 + 0.5*np.sum(new_model.mask[j,:]*new_model.phi[t,:])
        new_model.update_conventional_hyperparameters()
        new_model.update_expectations_etc()
    new_model.fit(tolerance = tolerance, max_iter = max_iter, min_iter = min_iter)
    return new_model


import numpy as np
import pandas as pd
import plotnine as p9
from copy import copy
from itertools import combinations, product
from scipy.stats import norm as norm_dist
from scipy.stats import beta as beta_dist
from scipy.stats import multinomial as multinomial_dist
from scipy.stats import gamma as gamma_dist
from scipy.stats import t as t_dist
from scipy.special import digamma, loggamma

def lin_reg(x, y, x_test = None, y_test = None, prior_pseudo_obs = 10):
    """
    Bayesian linear regression with a conjugate prior.
    
    Notes
    -----
    
    The regression weights (w) and precisions (xi) have a multi-mean normal-gamma prior with the following hyperparameters:
        m_w = 0
        Lambda_w = (prior_pseudo_obs/T)*I ***** CHECK *****
        alpha_xi = prior_pseudo_obs/2
        beta_xi = prior_pseudo_obs/2
        
        The fact that the prior alpha_xi = prior beta_xi implies that the prior distribution on xi
        has a mean of 1.
    """
    ##### FINISH #####


import numpy as np
import pandas as pd
import plotnine as p9
from itertools import product
from scipy.stats import beta as beta_dist
from scipy.stats import t as t_dist
from . import bayes_factors
from . import exp_families
from scipy.special import betaln

def _remove_nan(y, z):
    '''
    Convenience function to remove observations with missing data.
    '''
    is_nan = np.isnan(y)
    if np.sum(is_nan) > 0:
        y = y[~is_nan].copy()
        z = z[:, ~is_nan].copy()
    return (y, z)

def fit_y_normal(model,
                 y,
                 x_new = None,
                 y_new = None,
                 index = None,
                 profile_labels = None,
                 start_profile_labels_from1 = False,
                 prior_alpha = 1.0,
                 prior_beta = 1.0,
                 prior_m = 0.0,
                 prior_lambda = 1.0,
                 figure_size = [10, 12],
                 font_size = 11,
                 facet_var = 'variable',
                 ncol = 4):
    '''
    Analyze the relationship between dependent variables (y) -
    that are assumed to be normally distributed - and the profiles.
    This is done without changing the model's other variational parameters.
    '''
    ##### SET UP DATA ETC #####
    y_names = pd.Categorical(y.columns.values, categories = y.columns.values, ordered = True)
    if index is None:
        y = y.copy()
    else:
        y = y.iloc[index].copy().transpose()
    n_y = y.shape[1] # number of dependent variables
        
    bayes_factor_df = pd.DataFrame(1.0,
                                   columns = ['bf10', 'frac_bf10'],
                                   index = y_names) # empty dataframe for results
        
    ##### COMPUTE REGULAR, FRACTIONAL, AND POSTERIOR BAYES FACTORS #####
    for j in range(n_y):
        y_j = y[y_names[j]].values
        bayes_factor_df.loc[y_names[j], 'bf10'] = bayes_factors.normal_shared_var(y_j, model.z_hat_1hot, prior_m, prior_lambda, prior_alpha, prior_beta)
        bayes_factor_df.loc[y_names[j], 'frac_bf10'] = bayes_factors.frac_normal_shared_var(y_j, model.z_hat_1hot, prior_m, prior_lambda, prior_alpha, prior_beta)
        bayes_factor_df.loc[y_names[j], 'post_bf10'] = bayes_factors.post_normal_shared_var(y_j, model.z_hat_1hot, prior_m, prior_lambda, prior_alpha, prior_beta)
        
    ##### PAIRWISE COMPARISONS (FRACTIONAL BAYES FACTORS) TO PROFILE 0 #####
    pairwise_log10_frac_bf10 = pd.DataFrame(0.0,
                                            columns = ['profile ' + str(t + start_profile_labels_from1) for t in range(1, model.n_profiles)],
                                            index = y_names)
    for j in range(n_y):
        for t in range(1, model.n_profiles):
            in_profiles = np.isin(model.z_hat, [0, t])
            y_used = y.loc[in_profiles, y_names[j]].values # only include data from people in these profiles
            z_hat_used = model.z_hat_1hot[:, in_profiles][[0, t], :]
            pairwise_log10_frac_bf10.loc[y_names[j], 'profile ' + str(t + start_profile_labels_from1)] = np.log10(bayes_factors.frac_normal_shared_var(y_used, z_hat_used, prior_m, prior_lambda, prior_alpha, prior_beta))
    
    ##### POSTERIOR HYPERPARAMETERS #####
    post_hpar = {'m': np.zeros([n_y, model.n_profiles]), 'lambda': np.zeros([n_y, model.n_profiles]), 'alpha': np.zeros(n_y), 'beta': np.zeros(n_y)}
    v_mu = np.zeros([n_y, model.n_profiles])
    for j in range(n_y):
        (y_j, z_hat_j) = _remove_nan(y[y_names[j]].values, model.z_hat_1hot)
        dist = exp_families.multi_group_normal({'m': prior_m, 'lambda': prior_lambda, 'alpha': prior_alpha, 'beta': prior_beta}, model.n_profiles)
        dist.update(y_j, z_hat_j)
        post_hpar['m'][j,:] = dist.hpar['m']
        post_hpar['lambda'][j,:] = dist.hpar['lambda']
        post_hpar['alpha'][j] = dist.hpar['alpha']
        post_hpar['beta'][j] = dist.hpar['beta']
        v_mu[j, :] = dist.hpar['beta']/(dist.hpar['lambda']*(dist.hpar['alpha'] - 1))
    
    ##### COMPUTE THE COEFFICIENT OF DETERMINATION (R^2) #####
    r2 = pd.Series(0.0, index = y_names)
    for j in range(n_y):
        (y_j, z_hat_j) = _remove_nan(y[y_names[j]].values, model.z_hat_1hot)
        y_hat = np.inner(post_hpar['m'][j, :], z_hat_j.T)
        ss_residuals = np.sum((y_j - y_hat)**2)
        ss_total = np.sum((y_j - np.mean(y_j))**2)
        r2[y_names[j]] = 1 - ss_residuals/ss_total
    
    ##### MAKE PLOT #####
    plot_df = pd.DataFrame(product(range(n_y), range(model.n_profiles)),
                                columns = ['j', 'profile'])
    plot_df['variable'] = y_names[plot_df['j'].values]
    plot_df['mu'] = 0.0
    plot_df['v'] = 0.0
    hard_m_mu = np.zeros([n_y, model.n_profiles])
    hard_v_mu = np.zeros([n_y, model.n_profiles])
    for r in range(plot_df.shape[0]):
        plot_df.loc[r, 'mu'] = post_hpar['m'][plot_df.iloc[r]['j'].astype(int), plot_df.iloc[r]['profile'].astype(int)]
        plot_df.loc[r, 'v'] = v_mu[plot_df.iloc[r]['j'].astype(int), plot_df.iloc[r]['profile'].astype(int)]
    plot_df['mu_minus'] = plot_df['mu'] - 1.96*np.sqrt(plot_df['v'])
    plot_df['mu_plus'] = plot_df['mu'] + 1.96*np.sqrt(plot_df['v'])
    if start_profile_labels_from1:
        plot_df['profile'] += 1
    plot_df['profile'] = plot_df['profile'].astype('string')
    
    if facet_var == 'profile':
        if profile_labels is None:
            plot = p9.ggplot(plot_df, p9.aes(x = 'variable', y = 'mu', ymin = 'mu_minus', ymax = 'mu_plus'))
        else:
            plot = p9.ggplot(plot_df, p9.aes(shape = 'profile label', x = 'variable', y = 'mu', ymin = 'mu_minus', ymax = 'mu_plus'))
    elif facet_var == 'variable' or n_y == 1:
        if profile_labels is None:
            plot = p9.ggplot(plot_df, p9.aes(x = 'profile', y = 'mu', ymin = 'mu_minus', ymax = 'mu_plus'))
        else:
            plot = p9.ggplot(plot_df, p9.aes(shape = 'profile label', x = 'profile', y = 'mu', ymin = 'mu_minus', ymax = 'mu_plus'))

    plot += p9.geom_point()
    plot += p9.geom_errorbar()
    plot += p9.theme_classic(base_size = font_size)
    plot += p9.theme(figure_size = figure_size)
    if not profile_labels is None:
        plot += p9.scale_shape_manual(values = len(profile_labels)*["o"]) # this is just a hack to display the legend (we don't want different shapes) 
        
    if n_y > 1:
        if facet_var == 'profile':
            plot += p9.facet_wrap('profile', scales = 'free_x', ncol = ncol)
        elif facet_var == 'variable':
            plot += p9.facet_wrap('variable', scales = 'free', ncol = ncol)
    
    output = {'plot': plot, 'plot_df': plot_df, 'bayes_factors': bayes_factor_df, 'log10_bayes_factors': np.log10(bayes_factor_df), 'pairwise_log10_frac_bf10': pairwise_log10_frac_bf10, 'r2': r2, 'post_hpar': post_hpar}
    
#    if not y_new is None:
        # TOTALLY REVISE THIS
        ##### GET THE APPROXIMATE LOG POSTERIOR PREDICTIVE (p(y | x)) AND PREDICTED MEAN FOR NEW DATA #####
#        n_new = y_new.shape[0]
#        log_post_pred = pd.DataFrame(0.0, index = y_new.index, columns = y_new.columns)
#        post_pred_mean = pd.DataFrame(0.0, index = y_new.index, columns = y_new.columns)
#        phi_new = model.classify_new_data(x_new) # approximately p(z_new = t | x_new, old data)
#        for j in range(n_y):
#            var = y_new.columns[j]
#            post_pred_by_profile = np.zeros([model.T, n_new])
#            for t in range(model.T):
                # https://en.wikipedia.org/wiki/Conjugate_prior#When_likelihood_function_is_a_continuous_distribution ** DOUBLE CHECK THIS **
#                t_dist_sd = np.sqrt(beta[j]*(lamb[j, t] + 1)/(alpha[j]*lamb[j, t]))
#                post_pred_by_profile[t, :] = t_dist.pdf(y_new.loc[:, var],
#                                                        df = 2*alpha[j],
#                                                        loc = m[j, t],
#                                                        scale = t_dist_sd)
#            log_post_pred.loc[:, var] = np.log(np.sum(post_pred_by_profile*phi_new, axis = 0))
#            post_pred_mean.loc[:, var] = np.inner(m[j, :], phi_new.T)
            
        ##### COMPUTE THE PREDICTIVE COEFFICIENT OF DETERMINATION #####
#        predictive_r2 = pd.Series(0.0, index = y_names)
#        for var in y_names:
#            ss_residuals = np.sum((y_new[var] - post_pred_mean[var])**2)
#            ss_total = np.sum((y_new[var] - y_new[var].mean())**2)
#            predictive_r2[var] = 1 - ss_residuals/ss_total

#        output['predictive_r2'] = predictive_r2 # TOTALLY REVISE THIS
    return output
    
def fit_y_bernoulli(model, 
                    y,
                    index = None,
                    profile_labels = None,
                    start_profile_labels_from1 = False,
                    prior_a = 0.5,
                    prior_b = 0.5,
                    y_names = None, 
                    figure_size = [10, 8], 
                    font_size = 11, 
                    facet_var = 'variable', 
                    ncol = 4):
    '''
    Analyze the relationship between dependent variables (y) -
    that are assumed to be Bernoulli distributed - and the profiles.
    This is done without changing the model's other variational parameters.
    
    Notes
    -----
    We assume y_i | z_i = t ~ Bernoulli(psi_t).
    '''
    ##### SET UP DATA ETC #####
    y_names = pd.Categorical(y.columns.values, categories = y.columns.values, ordered = True)
    if index is None:
        y = y.copy()
    else:
        y = y.iloc[index].copy().transpose()
    n_y = y.shape[1] # number of dependent variables
    bayes_factor_df = pd.DataFrame(1.0,
                                   columns = ['bf10', 'frac_bf10'],
                                   index = y_names) # empty dataframe for results
    
    ##### COMPUTE BAYES FACTORS AND FRACTIONAL BAYES FACTORS #####
    for j in range(n_y):
        y_j = y[y_names[j]].values
        bayes_factor_df.loc[y_names[j], 'bf10'] = bayes_factors.bernoulli(y_j,
                                                                          model.z_hat_1hot,
                                                                          prior_a,
                                                                          prior_b)
        bayes_factor_df.loc[y_names[j], 'frac_bf10'] = bayes_factors.frac_bernoulli(y_j,
                                                                                    model.z_hat_1hot,
                                                                                    prior_a,
                                                                                    prior_b)
        bayes_factor_df.loc[y_names[j], 'post_bf10'] = bayes_factors.post_bernoulli(y_j,
                                                                                    model.z_hat_1hot,
                                                                                    prior_a,
                                                                                    prior_b)
    ##### POSTERIOR HYPERPARAMETERS #####
    post_hpar = dict()
    for j in range(n_y):
        (y_j, z_hat_j) = _remove_nan(y[y_names[j]].values, model.z_hat_1hot)
        dist = exp_families.multi_group_distribution(exp_families.beta_bernoulli, {'a': prior_a, 'b': prior_b}, model.n_profiles)
        dist.update(y_j, z_hat_j)
        post_hpar[y_names[j]] = dist.hpar_table()
        
    ##### PAIRWISE COMPARISONS (FRACTIONAL BAYES FACTORS) TO PROFILE 0 #####
    pairwise_log10_frac_bf10 = pd.DataFrame(0.0,
                                            columns = ['profile ' + str(t + start_profile_labels_from1) for t in range(1, model.n_profiles)],
                                            index = y_names)
    for j in range(n_y):
        for t in range(1, model.n_profiles):
            in_profiles = np.isin(model.z_hat, [0, t])
            y_used = y.loc[in_profiles, y_names[j]].values # only include data from people in these profiles
            z_hat_used = model.z_hat_1hot[:, in_profiles][[0, t], :]
            pairwise_log10_frac_bf10.loc[y_names[j], 'profile ' + str(t + start_profile_labels_from1)] = np.log10(bayes_factors.frac_bernoulli(y_used, z_hat_used, prior_a, prior_b))
    
    
    ##### COMPUTE THE BRIER SKILL SCORE (BSS) #####
    bss = pd.Series(0.0, index = y_names)
    for j in range(n_y):
        E_psi = post_hpar[y_names[j]]['a']/(post_hpar[y_names[j]]['a'] + post_hpar[y_names[j]]['b'])
        (y_j, z_hat_j) = _remove_nan(y[y_names[j]].values, model.z_hat_1hot)
        y_hat = np.inner(E_psi, z_hat_j.T)
        n_j = y_j.shape[0]
        Brier_skill = np.sum((y_j - y_hat)**2)/n_j
        reference_skill = np.sum((y_j - np.mean(y_j))**2)/n_j
        bss[y_names[j]] = 1 - Brier_skill/reference_skill
    
    ##### MAKE PLOT #####
    plot_df = pd.DataFrame(product(y_names, range(model.n_profiles)),
                           columns = ['variable', 'profile'])
    plot_df['E_psi'] = 0.0
    plot_df['lower'] = 0.0
    plot_df['upper'] = 0.0
    for j in range(n_y):
        rows_j = plot_df['variable'] == y_names[j]
        plot_df.loc[rows_j, 'E_psi'] = np.array(post_hpar[y_names[j]]['a']/(post_hpar[y_names[j]]['a'] + post_hpar[y_names[j]]['b']))
        plot_df.loc[rows_j, 'lower'] = beta_dist.ppf(0.05, post_hpar[y_names[j]]['a'], post_hpar[y_names[j]]['b'])
        plot_df.loc[rows_j, 'upper'] = beta_dist.ppf(0.95, post_hpar[y_names[j]]['a'], post_hpar[y_names[j]]['b'])
    if start_profile_labels_from1:
        plot_df['profile'] += 1
    plot_df['profile'] = plot_df['profile'].astype('string')
    if not profile_labels is None:
        plot_df['profile label'] = int(plot_df.shape[0]/len(profile_labels))*profile_labels

    if facet_var == 'profile':
        if profile_labels is None:
            plot = p9.ggplot(plot_df, p9.aes(x = 'variable', y = 'E_psi', ymin = 'lower', ymax = 'upper'))
        else:
            plot = p9.ggplot(plot_df, p9.aes(shape = 'profile label', x = 'variable', y = 'E_psi', ymin = 'lower', ymax = 'upper'))
    elif facet_var == 'variable' or n_y == 1:
        if profile_labels is None:
            plot = p9.ggplot(plot_df, p9.aes(x = 'profile', y = 'E_psi', ymin = 'lower', ymax = 'upper'))
        else:
            plot = p9.ggplot(plot_df, p9.aes(shape = 'profile label', x = 'profile', y = 'E_psi', ymin = 'lower', ymax = 'upper'))

        plot += p9.geom_point()
        plot += p9.geom_errorbar()
        plot += p9.theme_classic(base_size = font_size)
        plot += p9.theme(figure_size = figure_size)
        if not profile_labels is None:
            plot += p9.scale_shape_manual(values = len(profile_labels)*["o"]) # this is just a hack to display the legend (we don't want different shapes)

        if n_y > 1:
            if facet_var == 'profile':
                plot += p9.facet_wrap('profile', scales = 'free_x', ncol = ncol)
            elif facet_var == 'variable':
                plot += p9.facet_wrap('variable', scales = 'free', ncol = ncol)

    return {'plot': plot, 'plot_df': plot_df, 'bayes_factors': bayes_factor_df, 'log10_bayes_factors': np.log10(bayes_factor_df), 'pairwise_log10_frac_bf10': pairwise_log10_frac_bf10, 'Brier_skill_score': bss, 'post_hpar': post_hpar}